
// Copyright Aleksey Gurtovoy 2003-2008
//
// Distributed under the Boost Software License, Version 1.0. 
// (See accompanying file LICENSE_1_0.txt or copy at 
// http://www.boost.org/LICENSE_1_0.txt)
//
// See http://www.boost.org/libs/mpl for documentation.

// $Source$
// $Date: 2004-09-14 07:39:55 -0500 (Tue, 14 Sep 2004) $
// $Revision: 25084 $

#define BOOST_NO_CONFIG

#include <boost/preprocessor/config/config.hpp>

#define BOOST_NO_TEMPLATE_TEMPLATES
#define BOOST_MPL_CFG_NO_BIND_TEMPLATE
#define BOOST_MPL_CFG_NO_NONTYPE_TEMPLATE_PARTIAL_SPEC
#define BOOST_MPL_CFG_NO_DEFAULT_PARAMETERS_IN_NESTED_TEMPLATES
#define BOOST_MPL_CFG_BROKEN_DEFAULT_PARAMETERS_IN_NESTED_TEMPLATES
#define BOOST_MPL_CFG_BROKEN_OVERLOAD_RESOLUTION
#define BOOST_MPL_CFG_EXTENDED_TEMPLATE_PARAMETERS_MATCHING
#define BOOST_MPL_CFG_BCC_INTEGRAL_CONSTANTS
#define BOOST_MPL_CFG_NO_NESTED_FORWARDING
